@@ -6,7 +6,7 @@ import logging
 from typing import Final, Union
 import re
 import time
-import socket
+import asyncio
 from madvr.commands import ACKs, Footer, Commands, Enum, Connections
 from madvr.errors import AckError, RetryExceededError, HeartBeatError
 
@@ -59,8 +59,15 @@ class Madvr:
         self.outgoing_black_levels = ""
 
         # Sockets
-        self.client = None
-        self.notification_client = None
+        self.reader = None
+        self.writer = None
+        self.reader_lock = asyncio.Lock()
+        self.writer_lock = asyncio.Lock()
+
+        self.notification_reader = None
+        self.notification_writer = None
+        self.notification_reader_lock = asyncio.Lock()
+        self.notification_writer_lock = asyncio.Lock()
         self.is_closed = False
         # Envy does not have an are you on cmd, just assuming its on based on active connection
         self.is_on = False
@@ -68,7 +75,7 @@ class Madvr:
         self.command_read_timeout = 3
         self.logger.debug("Running in debug mode")
 
-    def _clear_attr(self) -> None:
+    async def _clear_attr(self) -> None:
         """
         Clear instance attr so HA doesn't report stale values
         """
@@ -98,38 +105,40 @@ class Madvr:
         self.outgoing_hdr_flag = False
         self.outgoing_black_levels = ""
 
-    def close_connection(self) -> None:
+    async def close_connection(self) -> None:
         """close the connection"""
-        try:
-            self.client.close()
-        except AttributeError:
-            # means its already closed
-            pass
-        
-        try:
-            self.notification_client.close()
-        except AttributeError:
-            # means its already closed
-            pass
+        async with self.writer_lock:
+            try:
+                self.writer.close()
+                await self.writer.wait_closed()
+            except AttributeError:
+                # means its already closed
+                pass
+        async with self.notification_writer_lock:
+            try:
+                self.notification_writer.close()
+                await self.notification_writer.wait_closed()
+            except AttributeError:
+                # means its already closed
+                pass
 
-        self.client = None
-        self.notification_client = None
+        self.reader = None
+        self.notification_reader = None
 
         self.is_closed = True
 
         # Clear attr
-        self._clear_attr()
+        await self._clear_attr()
 
-    def open_connection(self) -> None:
+    async def open_connection(self) -> None:
         """Open a connection"""
         self.logger.debug("Starting open connection")
-
         try:
-            self._reconnect()
+            await self._reconnect()
         except AckError as err:
             self.logger.error(err)
 
-    def _reconnect(self) -> None:
+    async def _reconnect(self) -> None:
         """
         Initiate keep-alive connection. This should handle any error and reconnect eventually.
 
@@ -141,22 +150,29 @@ class Madvr:
                 self.logger.info("Connecting to Envy: %s:%s", self.host, self.port)
 
                 # Command client
-                self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-                self.client.settimeout(self.command_read_timeout)
-                self.client.connect((self.host, self.port))
+                async with self.reader_lock, self.notification_reader_lock:
+                    self.reader, self.writer = await asyncio.wait_for(
+                        asyncio.open_connection(self.host, self.port),
+                        timeout=self.command_read_timeout,
+                    )
 
-                # Notifications client
-                self.notification_client = socket.socket(
-                    socket.AF_INET, socket.SOCK_STREAM
-                )
-                self.notification_client.settimeout(20)
-                self.notification_client.connect((self.host, self.port))
+                    # Notifications client
+                    (
+                        self.notification_reader,
+                        self.notification_writer,
+                    ) = await asyncio.wait_for(
+                        asyncio.open_connection(self.host, self.port), timeout=20
+                    )
 
                 # Test heartbeat
                 self.logger.debug("Handshaking")
 
                 # Make sure first message says WELCOME
-                msg_envy = self.client.recv(self.read_limit)
+                async with self.reader_lock:
+                    msg_envy = await asyncio.wait_for(
+                        self.reader.readline(),
+                        timeout=10,
+                    )
 
                 # Check if first 7 char match
                 if self.MADVR_OK != msg_envy[:7]:
@@ -166,7 +182,10 @@ class Madvr:
                     )
 
                 # Make sure first message says WELCOME for notifications
-                msg_envy = self.notification_client.recv(self.read_limit)
+                async with self.notification_reader_lock:
+                    msg_envy = await asyncio.wait_for(
+                        self.notification_reader.readline(), timeout=20
+                    )
 
                 # Check if first 7 char match
                 if self.MADVR_OK != msg_envy[:7]:
@@ -175,10 +194,10 @@ class Madvr:
                     )
                 self.logger.info("Waiting for envy to be available")
                 # envy needs some time to setup new connections
-                time.sleep(3)
+                await asyncio.sleep(3)
 
                 # handshake func
-                self._send_heartbeat()
+                await self._send_heartbeat()
 
                 self.logger.info("Connection established")
 
@@ -191,30 +210,29 @@ class Madvr:
                 self.logger.warning(
                     "Error sending heartbeat, retrying in %s seconds", 2
                 )
-                time.sleep(2)
+                await asyncio.sleep(2)
                 continue
 
             # includes conn refused
             # backoff to not spam HA
-            except socket.timeout:
-                self.logger.warning(
-                    "Connecting timeout, retrying in %s seconds", 2
-                )
-                time.sleep(2)
-                continue
-            except OSError as err:
-                self.logger.warning(
-                    "Connecting failed, retrying in %s seconds", 2
-                )
-                self.logger.debug(err)
-                time.sleep(2)
+            except asyncio.TimeoutError:
+                self.logger.warning("Connecting timeout, retrying in %s seconds", 2)
+                await asyncio.sleep(2)
                 continue
 
-    def _send_heartbeat(self) -> None:
+            # includes conn refused
+            # backoff to not spam HA
+            except OSError as err:
+                self.logger.warning("Connecting failed, retrying in %s seconds", 2)
+                self.logger.debug(err)
+                await asyncio.sleep(2)
+                continue
+
+    async def _send_heartbeat(self) -> None:
         """
         Send a heartbeat to keep connection open
 
-        You should wrap this in try with socket.error and socket.timeout exceptions
+        You should wrap this in try with OSError and asyncio.TimeoutError exceptions
 
         Raises HeartBeatError exception
         """
@@ -224,25 +242,22 @@ class Madvr:
             # confirm can send heartbeat, ready for commands
             self.logger.debug("Sending heartbeats")
 
-            self.client.send(self.HEARTBEAT)
+            async with self.writer_lock:
+                self.writer.write(self.HEARTBEAT)
+                await self.writer.drain()
 
             # Recv until we find ok or timeout
-            if not self._read_until_ok(self.client):
+            if not await self._read_until_ok(self.reader):
                 i += 1
                 continue
 
             # send heartbeat with notification and read ack and then regular client
             # ensure both work one after the next
-            self.notification_client.send(self.HEARTBEAT)
+            async with self.notification_writer_lock:
+                self.notification_writer.write(self.HEARTBEAT)
+                await self.notification_writer.drain()
 
-            if not self._read_until_ok(self.notification_client):
-                i += 1
-                continue
-
-            # send again on regular client to make sure both clients work
-            self.client.send(self.HEARTBEAT)
-
-            if not self._read_until_ok(self.client):
+            if not await self._read_until_ok(self.notification_reader):
                 i += 1
                 continue
 
@@ -252,7 +267,9 @@ class Madvr:
 
         raise HeartBeatError("Sending heartbeat fatal error")
 
-    def _construct_command(self, raw_command: Union[str, list]) -> tuple[bytes, bool, str]:
+    async def _construct_command(
+        self, raw_command: Union[str, list]
+    ) -> tuple[bytes, str]:
         """
         Transform commands into their byte values from the string value
 
@@ -260,15 +277,18 @@ class Madvr:
 
         Return:
             bytes: the value to send in bytes
-            bool: if its informational
             str: the 'msg' field in the Enum used to filter notifications
         """
-        self.logger.debug("raw_command: %s", raw_command)
+        self.logger.debug(
+            "raw_command: %s -- raw_command length: %s", raw_command, len(raw_command)
+        )
+
         skip_val = False
         # HA seems to always send commands as a list even if you set them as a str
 
         # This lets you use single cmds or something with val like KEYPRESS
         # If len is 1, then try to split, otherwise its just one word
+
         if len(raw_command) == 1:
             try:
                 # ['key_press, menu']
@@ -277,19 +297,22 @@ class Madvr:
                 value = raw_value.strip()
                 self.logger.debug("using command %s and value %s", command, value)
             # if valuerror it means theres just one command like PowerOff, so use that directly
-            except ValueError:
+            except ValueError as err:
+                self.logger.debug(err)
+                self.logger.debug("Using raw_command directly")
                 command = raw_command
                 skip_val = True
         # if there are more than two values, this is incorrect, error
         elif len(raw_command) > 2:
-            self.logger.error("More than two command values provided. Envy does not have more than 2 command values e.g KeyPress MENU")
+            self.logger.error(
+                "More than two command values provided. Envy does not have more than 2 command values e.g KeyPress MENU"
+            )
             raise NotImplementedError(f"Too many values provided {raw_command}")
         else:
             # else a command was provided as a proper list ['keypress', 'menu']
             # raw command will be a list of 2
             command, value = raw_command
 
-
         self.logger.debug("using command %s", command)
 
         # Check if command is implemented
@@ -297,7 +320,7 @@ class Madvr:
             raise NotImplementedError(f"Command not implemented: {command}")
 
         # construct the command with nested Enums
-        command_name, val, is_info = Commands[command].value
+        command_name, val, _ = Commands[command].value
 
         # if there is a value to process
         if not skip_val:
@@ -314,9 +337,9 @@ class Madvr:
 
         self.logger.debug("constructed command: %s", cmd)
 
-        return cmd, is_info.value, val
+        return cmd, val
 
-    def _read_until_ok(self, client: socket.socket) -> bool:
+    async def _read_until_ok(self, client: asyncio.StreamReader) -> bool:
         """
         Read the buffer until we get ok or timeout because a timeout means no more to read
         """
@@ -324,8 +347,9 @@ class Madvr:
         # exit loop if exceed maximum checks for ok or retries
         while True:
             try:
-                data = client.recv(self.read_limit)
-                self.logger.debug("_read_until_ok: data found: %s", data)
+                async with self.reader_lock, self.notification_reader_lock:
+                    data = await asyncio.wait_for(client.readline(), timeout=10)
+                    self.logger.debug("_read_until_ok: data found: %s", data)
 
                 if ACKs.reply.value not in data:
                     self.logger.debug("OK not found yet counter: %s", counter)
@@ -335,13 +359,14 @@ class Madvr:
                 self.logger.debug("OK found counter: %s", counter)
                 return True
 
-            except (socket.timeout, socket.error):
+            except asyncio.TimeoutError:
                 self.logger.debug("Timeout reading ack with counter: %s", counter)
                 return False
 
-    def send_command(self, command: Union[str, list]) -> str:
+    async def send_command(self, command: Union[str, list]) -> str:
         """
         send a given command same as the official madvr ones
+        To keep this simple, just send the command without reading response
 
         command: str - command to send like KeyPress, MENU
         Raises RetryExceededError
@@ -349,19 +374,19 @@ class Madvr:
 
         # Verify the command is supported
         try:
-            cmd, is_info, enum_type = self._construct_command(command)
+            cmd, enum_type = await self._construct_command(command)
         except NotImplementedError:
             return f"Command not found: {command}"
 
-        self.logger.debug("using values: %s %s %s", cmd, is_info, enum_type)
+        self.logger.debug("using values: %s %s", cmd, enum_type)
 
         # reconnect if client is not init or its off
-        if self.client is None or self.is_on is False:
+        if self.reader is None or self.is_on is False:
             # Don't reconnect if poweroff or standby because HA will complain
             if "PowerOff" in command or "Standby" in command:
                 return ""
             self.logger.debug("Connection lost - restarting connection")
-            self._reconnect()
+            await self._reconnect()
 
         # simple retry logic
         retry_count = 0
@@ -369,14 +394,11 @@ class Madvr:
         while retry_count < 5:
             # Send the command
             try:
-                self.client.send(cmd)
+                async with self.writer_lock:
+                    self.writer.write(cmd)
+                    await self.writer.drain()
 
-                if not self._read_until_ok(self.client):
-                    self.logger.debug("OK not found when reading response, retrying")
-                    retry_count += 1
-                    continue
-
-            except socket.timeout:
+            except asyncio.TimeoutError:
                 self.logger.debug("OK receipt timed out, retrying")
                 retry_count += 1
                 continue
@@ -387,91 +409,72 @@ class Madvr:
                 retry_count += 1
                 continue
 
-            # If its an info, get the rest of the info
-            try:
-                if is_info:
-                    # read response
-                    res = self.client.recv(self.read_limit)
-                    self.logger.debug("Response from info: %s", res)
-
-                    # TODO one day: whatever is not part of our command, write that to attr
-                    # e.g if we ask signal, and get notification for aspect, detect it and write that
-                    # so polling isn't required
-
-                    # process the output
-                    return self._process_info(res, enum_type["msg"].value)
-
-                return ""
-
-            except socket.timeout:
-                self.logger.debug("Ack receipt timed out, retrying")
-                retry_count += 1
-                continue
+            return ""
 
         # if we got here something went wrong
-        self.close_connection()
-        self._reconnect()
+        await self.close_connection()
+        await self._reconnect()
 
         return ""
 
-    def read_notifications(self, wait_forever: bool) -> None:
+    async def start_read_notifications(self, wait_forever: bool) -> None:
         """
         Listen for notifications. Meant to run as a background task
         wait_forever: bool -> if true, it will block forever. False useful for testing
         """
-        # TODO: should this be a second integration?
-        # Is there a way for HA to always poll in background?z
+        # reconnect if client is not init or its off
+        if self.notification_reader is None or self.is_on is False:
+            await self._reconnect()
 
         # Receive data in a loop
         i = 0
         while wait_forever or i < 5:
             try:
-                self.notification_client.sendall(self.HEARTBEAT)
-                data = self.notification_client.recv(self.read_limit)
-                time.sleep(1)
+                # send hearbeat
+                async with self.notification_writer_lock:
+                    self.notification_writer.write(self.HEARTBEAT)
+                    await self.notification_writer.drain()
+                
+                async with self.notification_reader_lock:
+                    data = await asyncio.wait_for(
+                        self.reader.readline(),
+                        timeout=self.command_read_timeout,
+                    )
+
+                asyncio.sleep(1)
+
                 i += 1
+
                 if data:
                     # process data which will get added as class attr
-                    self._process_notifications(data)
+                    await self._process_notifications(data)
                 else:
                     # just wait forever for data
                     # send heartbeat keep conn open
-                    self.notification_client.sendall(self.HEARTBEAT)
+                    async with self.notification_writer_lock:
+                        self.notification_writer.write(self.HEARTBEAT)
+                        await self.notification_writer.drain()
                     continue
-            except socket.timeout:
+            except asyncio.TimeoutError:
                 self.logger.debug("Connection timed out")
-                self.notification_client.sendall(self.HEARTBEAT)
+                async with self.notification_writer_lock:
+                    self.notification_writer.write(self.HEARTBEAT)
+                    await self.notification_writer.drain()
                 continue
-            except socket.error:
+            except OSError:
                 self.logger.debug("Connection error")
-                self.notification_client.sendall(self.HEARTBEAT)
+                async with self.notification_writer_lock:
+                    self.notification_writer.write(self.HEARTBEAT)
+                    await self.notification_writer.drain()
                 continue
+            except AttributeError:
+                await self._reconnect()
 
-    def poll_status(self) -> None:
-        """
-        Poll the status for attributes and write them to state
-        """
-        try:
-            # send heartbeat so it doesnt close our connection
-            self._send_heartbeat()
-            # Get incoming signal info
-            for cmd in [
-                "GetIncomingSignalInfo",
-                "GetAspectRatio",
-                "GetTemperatures",
-                "GetOutgoingSignalInfo",
-            ]:
-                res = self.send_command(cmd)
-                self.logger.debug("poll_status resp: %s", res)
-                self._process_notifications(res)
-        except (socket.timeout, socket.error, HeartBeatError) as err:
-            self.logger.error("Error getting update: %s", err)
-
-    def _process_notifications(self, input_data: Union[bytes, str]) -> None:
+    async def _process_notifications(self, input_data: Union[bytes, str]) -> None:
         """
         Process arbitrary stream of notifications and set them as instance attr
         """
-
+        # This code constructs a dict for all values processed
         self.logger.debug("Processing data for %s", input_data)
         try:
             if isinstance(input_data, bytes):
@@ -535,7 +538,7 @@ class Madvr:
             self.outgoing_colorimetry = outgoing_signal_info[6]
             self.outgoing_black_levels = outgoing_signal_info[7]
 
-    def power_off(self, standby=False) -> str:
+    async def power_off(self, standby=False) -> str:
         """
         turn off madvr it must have a render thread active at the moment
         once it is off, you can't turn it back on unless standby=True
@@ -546,8 +549,8 @@ class Madvr:
         # dont do anything if its off besides mark it off
         # sending command will open connection
         try:
-            res = self.send_command("Standby" if standby else "PowerOff")
-            self.close_connection()
+            res = await self.send_command("Standby" if standby else "PowerOff")
+            await self.close_connection()
             self.is_on = False
 
             return res
@@ -555,24 +558,6 @@ class Madvr:
         except RetryExceededError:
             return "Retries Exceeded"
 
-    def _process_info(self, input_data: bytes, filter_str: str) -> str:
-        """
-        Process info given input and a filter str to only return the thing we want
-        e.g for IncomingSignalInfo
-        b"Ok\r\nIncomingSignalInfo 3840x2160 23.976p 2D 422
-        10bit HDR10 2020 TV 16:9\r\nAspect Ratio ETC ETC\r\n"
-        turns into -> IncomingSignalInfo 3840x2160 23.976p 2D 422 10bit HDR10 2020 TV 16:9
-
-        This is used by _process_notifications to turn it into a dict, add to instance attr
-        """
-
-        lines = input_data.decode().split("\r\n")
-        for line in lines:
-            if filter_str in line:
-                return line
-
-        return ""
-
     def print_commands(self) -> str:
         """
         Print out all supported commands
